/*
 *  rt_start.C - main procedures for ray tracer
 *
 *
 *  Created by Sriram Vaidhyanathan on Thu Feb 26 2004.
 *  15-462 Computer Graphics Spring 2004 Programming Assignment 3
 *
 */

#include <iostream>
#ifdef TARGET_OS_MAC
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#endif
#include <pic.h>

#include "rt_parse.h"

using namespace std;

char *filename = 0;

/* different display modes */
#define MODE_DISPLAY   1
#define MODE_JPEG      2

/* Set to display, not jpeg */
int mode = MODE_DISPLAY;
rt_parse *file_parser=NULL;

/**********************************************************/
/* You may want to reduce these for
 * debugging purposes - cutting each dimension in
 * half, e.g. 320x240 cuts the rays traced to a
 * quarter of their original value. Set these back to
 * 640x480 for the final submission, but your code should
 * be written to handle any dimensions for the image.     */
/**********************************************************/
#define WIDTH    640
#define HEIGHT   480

/* Buffer to write pixels to */
unsigned char buffer[HEIGHT * WIDTH * 3];

/* Some helpful functions in here */
class plot_pixel_types
{
public:
  static void plot_pixel_display(int x, int y, unsigned char r,
				 unsigned char g, unsigned char b);
  static void plot_pixel_jpeg(int x, int y, unsigned char r,
			      unsigned char g, unsigned char b);
  static void plot_pixel(int x, int y, unsigned char r,
			 unsigned char g, unsigned char b);
};

/* Write out pixels to display window */
void plot_pixel_types::plot_pixel_display(int x, int y, unsigned char r,
					  unsigned char g, unsigned char b)
{
  glBegin(GL_POINTS);
  {
    glColor3f(((float)r)/256.f, ((float)g)/256.f, ((float)b)/256.f);
    glVertex2i(x, y);
  }
  glEnd();
  
  glFlush();
}

/* Write out pixels to jpeg buffer */
void plot_pixel_types::plot_pixel_jpeg(int x, int y, unsigned char r,
				       unsigned char g, unsigned char b)
{
  buffer[(HEIGHT-y-1) * 3 * WIDTH + x*3] = r;
  buffer[(HEIGHT-y-1) * 3 * WIDTH + x*3 + 1] = g;
  buffer[(HEIGHT-y-1) * 3 * WIDTH + x*3 + 2] = b;
}

/* Combines both functions here */
void plot_pixel_types::plot_pixel(int x, int y, unsigned char r,
				  unsigned char g, unsigned char b)
{
  plot_pixel_display(x, y, r, g, b);
  if(mode == MODE_JPEG)
    plot_pixel_jpeg(x, y, r, g, b);
}

/******************************************************/
/* You will need to modify this function as required. */
/******************************************************/
void draw_scene()
{
  unsigned int x, y;

  /* Just some simple temporary output here */
  for(x = 0; x < WIDTH; x++)
    for(y = 0; y < HEIGHT; y++)
      plot_pixel_types::plot_pixel(x, y, x%256, y%256, (x+y)%256);
}

/* Write jpeg buffer to file */
void save_jpg()
{
  Pic *in = NULL;
  in = pic_alloc(640, 480, 3, NULL);

  cout<<"Saving JPEG file: %s\n"<<filename;
  memcpy(in->pix, buffer, 3 * WIDTH * HEIGHT);
  
  if (jpeg_write(filename, in))
    cout<<"File saved successfully!\n";
  else
    cout<<"Error in saving\n";

  pic_free(in);
}

void display()
{
  /* Nothing in here */
}

void init()
{
  glMatrixMode(GL_PROJECTION);
  glOrtho(0, WIDTH, 0, HEIGHT, 1, -1);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glClearColor(0, 0, 0, 0);
  glClear(GL_COLOR_BUFFER_BIT);
}

void idle()
{
  /* Make sure this gets drawn only once */
  static int once = 0;

  if(!once)
    {
      /* modify this to turn off the display */
      draw_scene();
      if(mode == MODE_JPEG)
	save_jpg();
    }

  once = 1;
}

/* Program control starts in here */
int main(int argc, char **argv)
{
  if(argc < 2 || argc > 3)
    {  
      cout<<"Usage: "<<argv[0]<<" <scenefile> [jpegname]\n";
      exit(0);
    }
  
  if(argc == 3)
    {
      mode = MODE_JPEG;
      filename = argv[2];
    }
  else if(argc == 2)
    mode = MODE_DISPLAY;

  /**********************************************/
  /* Make the calls to parse the .ray file here */
  /**********************************************/
  
  file_parser = new rt_parse(argv[1]);
  
  glutInit(&argc,argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_SINGLE);
  glutInitWindowPosition(0,0);
  glutInitWindowSize(WIDTH,HEIGHT);
  glutCreateWindow("15-462 Programming Assignment 3: Ray Tracer");
  glutDisplayFunc(display);
  glutIdleFunc(idle);
  init();
  draw_scene();
  glutMainLoop();
}
